package org.dasein.cloud.test;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.Locale;
import java.util.UUID;

import org.apache.commons.httpclient.HttpClient;
import org.apache.commons.httpclient.HttpException;
import org.apache.commons.httpclient.HttpStatus;
import org.apache.commons.httpclient.methods.GetMethod;
import org.apache.commons.httpclient.params.HttpClientParams;
import org.apache.commons.httpclient.params.HttpMethodParams;
import org.dasein.cloud.CloudException;
import org.dasein.cloud.CloudProvider;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.services.cdn.ContentDistributionServices;
import org.dasein.cloud.services.cdn.Distribution;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

public class CDNTestCase extends BaseTestCase {
    private String        distributionToDelete = null;
    private CloudProvider provider             = null;
    private String        testFileName         = null;
    private String        testDirectory        = null;
    private Distribution  testDistribution     = null;
    private String        testDistributionId   = null;

    public CDNTestCase(String name) { super(name); }

    static private final String LINE_ONE = "1: Test";
    static private final String LINE_TWO = "2: Done.";
    
    private File createTestFile(String name) {
        try {
            PrintWriter writer = new PrintWriter(new BufferedWriter(new OutputStreamWriter(new FileOutputStream(name))));
            
            writer.println(LINE_ONE);
            writer.println(LINE_TWO);
            writer.flush();
            writer.close();
            return new File(name);
        }
        catch( Throwable t ) {
            return null;
        }
    }
    
    private void download(boolean retry) throws HttpException, IOException {
        GetMethod get = new GetMethod("http://" + testDistribution.getDnsName() + "/" + testFileName);
        
        try {
            HttpClientParams httpClientParams = new HttpClientParams();
            HttpClient client;
            
            httpClientParams.setParameter(HttpMethodParams.USER_AGENT, "Dasein Cloud");
            client = new HttpClient(httpClientParams);
            
            int status = client.executeMethod(get);
            
            if( status != HttpStatus.SC_OK ) {
                if( retry ) {
                    try { Thread.sleep(60000L); }
                    catch( InterruptedException e ) { }
                    download(false);
                }
                else {
                    fail("File was not found");
                }
            }
        }
        finally {
            get.releaseConnection();
        }
    }  
    
    @Before
    public void setUp() throws InternalException, CloudException, InstantiationException, IllegalAccessException {
        String name = getName();
        
        provider = getProvider();
        provider.connect(getTestContext());
        if( name.equals("testCreateDistribution") ) {
            testDirectory = provider.getStorageServices().createDirectory("dsncdn", true);
        }
        else if( name.equals("testGetDistribution") ) {
            for( Distribution d : provider.getCdnServices().list() ) {
                if( d.isActive() ) {
                    testDistributionId = d.getProviderDistributionId();
                }
            }
            if( testDistributionId == null ) {
                String dir = provider.getStorageServices().createDirectory("dsncdnsecondary", true);
                
                testDistributionId = provider.getCdnServices().create(dir, "Get CDN Test", true, System.currentTimeMillis() + "a.dasein.org");
                distributionToDelete = testDistributionId;
            }
        }
        else if( name.equals("testDistributionContent") ) {
            for( Distribution d : provider.getCdnServices().list() ) {
                if( d.isActive() ) {
                    testDistributionId = d.getProviderDistributionId();
                }
            }
            if( testDistributionId == null ) {
                String dir = provider.getStorageServices().createDirectory("dsncdnthird", true);
                
                testDistributionId = provider.getCdnServices().create(dir, "CDN Test Content", true, System.currentTimeMillis() + "b.dasein.org");
                distributionToDelete = testDistributionId;
            }            
        }
        else if( name.equals("testUpdateDistribution") ) {
            String dir = provider.getStorageServices().createDirectory("dsncdnfourth", true);
            
            testDistributionId = provider.getCdnServices().create(dir, "CDN Test Update", true, System.currentTimeMillis() + "c.dasein.org");
            distributionToDelete = testDistributionId;   
        }
        else if( name.equals("testDeleteDistribution") ) {
            String dir = provider.getStorageServices().createDirectory("dsncdnfive", true);
            
            testDistributionId = provider.getCdnServices().create(dir, "CDN Test Delete", true, System.currentTimeMillis() + "d.dasein.org");
            distributionToDelete = testDistributionId;   
        }   
        else if( name.equals("testDownloadFile") ) {
            String dir = provider.getStorageServices().createDirectory("dsncdnsix", true);
            
            testDistributionId = provider.getCdnServices().create(dir, name, true, System.currentTimeMillis() + "e.dasein.org");
            distributionToDelete = testDistributionId;
            testDistribution = provider.getCdnServices().getDistribution(testDistributionId);
            File sourceFile = createTestFile("dsntest.txt");
            
            provider.getStorageServices().upload(sourceFile, dir, "dsntest.txt", false, null);
        }
    }
    
    @After
    public void tearDown() {
        try {
            if( distributionToDelete != null ) {
                getProvider().getCdnServices().delete(distributionToDelete);
            }
        }
        catch( Throwable t ) {
            t.printStackTrace();
        }
        try {
            if( provider != null ) {
                provider.close();
            }
        }
        catch( Throwable t ) {
            t.printStackTrace();
        }
    }
    
    @Test
    public void testCreateDistribution() throws InternalException, CloudException {
        String id = provider.getCdnServices().create(testDirectory, "CDN Test", true, System.currentTimeMillis() + ".dasein.org");
        
        assertNotNull("No distribution was created", id);
        assertNotNull("No distribution was created", provider.getCdnServices().getDistribution(id));
    }
    
    @Test
    public void testDeleteDistribution() throws InternalException, CloudException {
        provider.getCdnServices().delete(testDistributionId);
        try { Thread.sleep(5000L); }
        catch( InterruptedException e ) { }
        Distribution d = provider.getCdnServices().getDistribution(testDistributionId);
        
        assertTrue("Distribution was not deleted", d == null || !d.isActive());
        distributionToDelete = null;
    }
    
    @Test
    public void testDistributionContent() throws CloudException, InternalException {
        Distribution d = provider.getCdnServices().getDistribution(testDistributionId);
        
        assertNotNull("No matching distribution", d);
        assertNotNull("No name exists for distribution", d.getName());
        assertEquals("Distribution ID does not match", testDistributionId, d.getProviderDistributionId());
    }
    
    @Test
    public void testDownloadFile() throws CloudException, InternalException, HttpException, IOException {
        download(true);
    }
    
    @Test
    public void testGetBogusDistribution() throws CloudException, InternalException {
        Distribution d = provider.getCdnServices().getDistribution(UUID.randomUUID().toString());
        
        assertNull("Distribution was incorrectly found", d);
    }
    
    @Test
    public void testGetDistribution() throws CloudException, InternalException {
        Distribution d = provider.getCdnServices().getDistribution(testDistributionId);
        
        assertNotNull("Distribution was not found", d);
    }
    
    @Test
    public void testListDistributions() throws CloudException, InternalException {
        assertNotNull("No distributions were provided", provider.getCdnServices().list());
    }
    
    @Test
    public void testMetaData() throws CloudException, InternalException {
        ContentDistributionServices services = provider.getCdnServices();
        
        assertNotNull("Must provider a provider name", services.getProviderTermForDistribution(Locale.getDefault()));
    }
    
    @Test
    public void testSubscription() throws CloudException, InternalException {
        // CURRENTLY A NO-OP
    }
    
    @Test
    public void testUpdateDistribution() throws CloudException, InternalException {
        provider.getCdnServices().update(testDistributionId, "Renamed Test", true, System.currentTimeMillis() + "c.dasein.org");
        try { Thread.sleep(5000L); }
        catch( InterruptedException e ) { }
        Distribution d = provider.getCdnServices().getDistribution(testDistributionId);
        
        assertEquals("Renamed Test", d.getName());
    }
}
