package org.dasein.cloud.test;

import java.util.Locale;
import java.util.UUID;

import org.dasein.cloud.CloudException;
import org.dasein.cloud.CloudProvider;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.services.firewall.Firewall;
import org.dasein.cloud.services.firewall.FirewallRule;
import org.dasein.cloud.services.firewall.Protocol;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

public class FirewallTestCase extends BaseTestCase {
    public FirewallTestCase(String name) { super(name); }

    private String        firewallToDelete = null;
    private CloudProvider provider         = null;
    private String        testFirewall     = null;
    
    @Before
    @Override
    public void setUp() throws InstantiationException, IllegalAccessException, CloudException, InternalException {
        provider = getProvider();
        provider.connect(getTestContext());
        String name = getName();
        
        if( name.equals("testAddFirewallRule") || name.equals("testDeleteFirewall") || name.equals("testRemoveFirewallRule") ) {
            testFirewall = provider.getFirewallServices().create("dsntest-" + System.currentTimeMillis(), "Firewall for testing.");
            firewallToDelete = testFirewall;
        }
        else if( name.equals("testFirewallContent") || name.equals("testGetFirewall") || name.equals("testListFirewallRules") ) {
            for( Firewall fw : provider.getFirewallServices().list() ) {
                if( fw.isActive() ) {
                    testFirewall = fw.getProviderFirewallId();
                    break;
                }
            }
            if( testFirewall == null ) {
                testFirewall = provider.getFirewallServices().create("dsntest-" + System.currentTimeMillis(), "Firewall for testing.");
                firewallToDelete = testFirewall;                
            }
        }
    }

    @After
    @Override
    public void tearDown() {
        try {
            if( firewallToDelete != null ) {
                provider.getFirewallServices().delete(firewallToDelete);
                firewallToDelete = null;
            }
        }
        catch( Throwable t ) {
            t.printStackTrace();
        }
        try {
            if( provider != null ) {
                provider.close();
            }
        }
        catch( Throwable t ) {
            t.printStackTrace();
        }        
    }

    @Test
    public void testAddFirewallRule() throws CloudException, InternalException {
        provider.getFirewallServices().authorize(testFirewall, "216.243.161.193/32", Protocol.TCP, 80, 80);
        try { Thread.sleep(5000L); }
        catch( InterruptedException e ) { }
        boolean added = false;
        
        for( FirewallRule rule : provider.getFirewallServices().getRules(testFirewall) ) {
            if( rule.getCidr().equals("216.243.161.193/32") ) {
                if( rule.getProtocol().equals(Protocol.TCP) ) {
                    if( rule.getStartPort() == 80 ) {
                        if( rule.getEndPort() == 80 ) {
                            added = true;
                        }
                    }
                }
            }
        }
        assertTrue("Firewall rule was not added", added);
    }
    
    @Test
    public void testCreateFirewall() throws InternalException, CloudException {
        testFirewall = provider.getFirewallServices().create("dsntest-create-" + System.currentTimeMillis(), "Dasein firewall create test");
        try { Thread.sleep(5000L); }
        catch( InterruptedException e ) { }
        Firewall fw = provider.getFirewallServices().getFirewall(testFirewall);
        
        assertNotNull("Did not create a firewall", fw);
        firewallToDelete = testFirewall;
    }
    
    @Test
    public void testDeleteFirewall() throws InternalException, CloudException {
        provider.getFirewallServices().delete(testFirewall);
        try { Thread.sleep(5000L); }
        catch( InterruptedException e ) { }
        Firewall fw = provider.getFirewallServices().getFirewall(testFirewall);
        
        assertNull("Deleted firewall still exists", fw);
        firewallToDelete = null;
    }
    
    @Test
    public void testFirewallContent() throws InternalException, CloudException {
        Firewall fw = provider.getFirewallServices().getFirewall(testFirewall);
        
        assertNotNull("A name must be set", fw.getName());
        assertNotNull("A description must be set", fw.getDescription());
        assertEquals("Firewall region must match request region", provider.getContext().getRegionId(), fw.getRegionId());
    }
    
    @Test
    public void testGetBogusFirewall() throws InternalException, CloudException {
        Firewall fw = provider.getFirewallServices().getFirewall(UUID.randomUUID().toString());
        
        assertNull("Found a matching firewall for bogus ID", fw);
    }
    
    @Test 
    public void testGetFirewall() throws InternalException, CloudException {
        Firewall fw = provider.getFirewallServices().getFirewall(testFirewall);
        
        assertNotNull("Could not find target firewall", fw);
        assertEquals("ID does not match returned firewall", testFirewall, fw.getProviderFirewallId());
    }
    
    @Test
    public void testListFirewalls() throws InternalException, CloudException {
        assertNotNull("Firewall list may not be null", provider.getFirewallServices().list());
    }
    
    @Test
    public void testListFirewallRules() throws InternalException, CloudException {
        assertNotNull("Firewall rules may not be null", provider.getFirewallServices().getRules(testFirewall));
    }
    
    @Test
    public void testMetaData() {
        assertNotNull("Must provide a provider name for firewalls", provider.getFirewallServices().getProviderTermForFirewall(Locale.getDefault()));
    }
    
    @Test
    public void testRemoveFirewallRule() throws CloudException, InternalException {
        provider.getFirewallServices().revoke(testFirewall, "216.243.161.193/32", Protocol.TCP, 80, 80);
        try { Thread.sleep(5000L); }
        catch( InterruptedException e ) { }
        boolean removed = true;
        
        for( FirewallRule rule : provider.getFirewallServices().getRules(testFirewall) ) {
            if( rule.getCidr().equals("216.243.161.193/32") ) {
                if( rule.getProtocol().equals(Protocol.TCP) ) {
                    if( rule.getStartPort() == 80 ) {
                        if( rule.getEndPort() == 80 ) {
                            removed = false;
                        }
                    }
                }
            }
        }
        assertTrue("Firewall rule was not removed", removed);
    }
    
    @Test
    public void testSubscription() {
        // Currently a NO-OP
    }
}
