package org.dasein.cloud.test;

import java.io.IOException;
import java.util.UUID;

import org.dasein.cloud.CloudException;
import org.dasein.cloud.CloudProvider;
import org.dasein.cloud.InternalException;
import org.dasein.cloud.OperationNotSupportedException;
import org.dasein.cloud.services.address.AddressServices;
import org.dasein.cloud.services.address.AddressType;
import org.dasein.cloud.services.address.IpAddress;
import org.dasein.cloud.services.address.IpForwardingRule;
import org.dasein.cloud.services.firewall.Protocol;
import org.junit.After;
import org.junit.Before;
import org.junit.Test;

public class IpAddressTestCase extends BaseTestCase {
    private String        addressToRelease = null;
    private CloudProvider provider         = null;
    private String        testAddress      = null;
    private String        testRuleId       = null;
    private String        vmToKill         = null;
    
    public IpAddressTestCase(String name) { super(name); }
    
    // TODO: IPv6 support
    private void assertValidAddress(String address) {
        assertNotNull("IP address value is empty", address);
        String[] parts = address.split("\\.");
        
        assertEquals("Address " + address + " is not a valid IPv4 address", 4, parts.length);
        for( int i=0; i<parts.length; i++ ) {
            try {
                int x = Integer.parseInt(parts[i]);
                
                assertTrue("Address " + address + " is not a valid IPv4 address", x > -1 && x < 256);
            }
            catch( NumberFormatException e ) {
                fail("Address " + address + " is not a valid IPv4 address");
            }
        }
    }
    
    private void initAddress() throws InternalException, CloudException {
        String name = getName();
        
        if( name.startsWith("testReleasePublicIpFromPool") ) {
            if( provider.getAddressServices().isRequestable(AddressType.PUBLIC) ) {
                addressToRelease = provider.getAddressServices().request(AddressType.PUBLIC);
                testAddress = addressToRelease;
            }
        }
        else if( name.startsWith("testReleasePrivateIpFromPool") ) {
            if( provider.getAddressServices().isRequestable(AddressType.PRIVATE) ) {
                addressToRelease = provider.getAddressServices().request(AddressType.PRIVATE);
                testAddress = addressToRelease;
            }
        }  
        else if( name.startsWith("testReleasePublicIpFromServer") ) {
            if( provider.getAddressServices().isAssigned(AddressType.PUBLIC) ) {
                for( IpAddress addr : provider.getAddressServices().listPublicIpPool(true) ) {
                    testAddress = addr.getProviderIpAddressId();
                    break;
                }
                if( testAddress == null ) {
                    if( provider.getAddressServices().isRequestable(AddressType.PUBLIC) ) {
                        addressToRelease = provider.getAddressServices().request(AddressType.PUBLIC);
                        testAddress = addressToRelease;
                    }
                }
                if( testAddress == null ) {
                    throw new CloudException("Unable to identify a free IP address for test execution.");
                } 
            }
        } 
        else if( name.startsWith("testReleasePrivateIpFromServer") ) {
            if( provider.getAddressServices().isAssigned(AddressType.PRIVATE) ) {
                for( IpAddress addr : provider.getAddressServices().listPublicIpPool(true) ) {
                    testAddress = addr.getProviderIpAddressId();
                    break;
                }
                if( testAddress == null ) {
                    if( provider.getAddressServices().isRequestable(AddressType.PUBLIC) ) {
                        addressToRelease = provider.getAddressServices().request(AddressType.PUBLIC);
                        testAddress = addressToRelease;
                    }
                }
                if( testAddress == null ) {
                    throw new CloudException("Unable to identify a free IP address for test execution.");
                } 
            }
        } 
        else if( name.equals("testAssignPrivateIp") ) {
            if( provider.getAddressServices().isAssigned(AddressType.PRIVATE) ) {
                if( provider.getAddressServices().isRequestable(AddressType.PRIVATE) ) {
                    addressToRelease = provider.getAddressServices().request(AddressType.PRIVATE);
                    testAddress = addressToRelease;
                }
                else {
                    for( IpAddress addr : provider.getAddressServices().listPrivateIpPool(true) ) {
                        testAddress = addr.getProviderIpAddressId();
                        break;
                    }
                    if( testAddress == null ) {
                        throw new CloudException("Unable to identify a free IP address for test execution.");
                    }
                }
            }
        }
        else if( name.equals("testAssignPublicIp") || name.equals("testForward") || name.equals("testStopForward") || name.equals("testStopBogusForward") || name.equals("testListForwardRules") ) {
            if( provider.getAddressServices().isAssigned(AddressType.PUBLIC) ) {
                if( provider.getAddressServices().isRequestable(AddressType.PUBLIC) ) {
                    addressToRelease = provider.getAddressServices().request(AddressType.PUBLIC);
                    testAddress = addressToRelease;
                }
                else {
                    for( IpAddress addr : provider.getAddressServices().listPublicIpPool(true) ) {
                        testAddress = addr.getProviderIpAddressId();
                        break;
                    }
                    if( testAddress == null ) {
                        throw new CloudException("Unable to identify a free IP address for test execution.");
                    }
                }
            }
        }
        else if( name.equals("testGetAddress") || name.equals("testAddressContent") ) {
            for( IpAddress addr : provider.getAddressServices().listPublicIpPool(false) ) {
                testAddress = addr.getProviderIpAddressId();
                break;
            }
            if( testAddress == null ) {
                for( IpAddress addr : provider.getAddressServices().listPrivateIpPool(false) ) {
                    testAddress = addr.getProviderIpAddressId();
                    break;
                }
            }            
            if( testAddress == null && provider.getAddressServices().isRequestable(AddressType.PUBLIC) ) {
                addressToRelease = provider.getAddressServices().request(AddressType.PUBLIC);
                testAddress = addressToRelease;
            }
            if( testAddress == null && provider.getAddressServices().isRequestable(AddressType.PRIVATE) ) {
                addressToRelease = provider.getAddressServices().request(AddressType.PRIVATE);
                testAddress = addressToRelease;
            }
        }
    }
    
    private void initVm() throws InternalException, CloudException {
        AddressServices services = provider.getAddressServices();
        String name = getName();
        
        if( name.equals("testAssignPrivateIp") || name.equals("testReleasePrivateIpFromServer") ) {
            if( services.isAssigned(AddressType.PRIVATE) ) {
                vmToKill = launch(provider);
            }
        }
        else if( name.equals("testAssignPublicIp") || name.equals("testReleasePublicIpFromServer")  ) {
            if( services.isAssigned(AddressType.PRIVATE) ) {
                vmToKill = launch(provider);
            }
        }
        else if( name.equals("testForward") || name.equals("testStopForward") ) {
            if( services.isForwarding() ) {
                vmToKill = launch(provider);
            }
        }
    }
    
    @Before
    @Override
    public void setUp() throws CloudException, InternalException, InstantiationException, IllegalAccessException, IOException {
        provider = getProvider();
        provider.connect(getTestContext());
        initAddress();
        initVm();
        if( getName().equals("testStopForward") ) {
            AddressServices services = provider.getAddressServices();
            
            if( services.isForwarding() ) {
                testRuleId = services.forward(testAddress, 9321, Protocol.TCP, 22, vmToKill);
            }
        }
        else if( getName().equals("testReleasePrivateIpFromServer") ) {
            AddressServices services = provider.getAddressServices();

            if( services.isAssigned(AddressType.PRIVATE) ) {
                services.assign(testAddress, vmToKill);
            }
        }
        else if( getName().equals("testReleasePublicIpFromServer") ) {
            AddressServices services = provider.getAddressServices();

            if( services.isAssigned(AddressType.PUBLIC) ) {
                services.assign(testAddress, vmToKill);
            }
        }        
    }
    
    @After
    @Override
    public void tearDown() throws CloudException, InternalException {
        try {
            if( testRuleId != null ) {
                provider.getAddressServices().stopForward(testRuleId);
                testRuleId = null;
            }
        }
        catch( Throwable ignore ) {
            ignore.printStackTrace();
        }
        try {
            if( addressToRelease != null ) {
                provider.getAddressServices().releaseFromPool(addressToRelease);
                addressToRelease = null;
            }
        }
        catch( Throwable ignore ) {
            ignore.printStackTrace();
        }
        try {
            if( vmToKill != null ) {
                provider.getServerServices().stop(vmToKill);
                vmToKill = null;
            }
        }
        catch( Throwable ignore ) {
            ignore.printStackTrace();
        }
        try {
            if( provider != null ) {
                provider.close();
                provider = null;
            }
        }
        catch( Throwable ignore ) {
            ignore.printStackTrace();
        }
    }
    
    @Test
    public void testAddressContent() throws CloudException, InternalException {
        if( testAddress == null ) {
            return;
        }
        AddressServices services = provider.getAddressServices();

        IpAddress addr = services.getIpAddress(testAddress);
        
        assertNotNull("Address not found: " + testAddress, addr);
        assertEquals("Address does not match target", testAddress, addr.getProviderIpAddressId());
        assertValidAddress(addr.getAddress());
        assertNotNull("An address must be either public or private", addr.getAddressType());
        assertEquals("Region does not match", provider.getContext().getRegionId(), addr.getRegionId());
    }
    
    @Test
    public void testAssignPrivateIp() throws CloudException, InternalException {
        AddressServices services = provider.getAddressServices();
        
        if( services.isAssigned(AddressType.PRIVATE) ) {
            services.assign(testAddress, vmToKill);
            try { Thread.sleep(5000L); }
            catch( InterruptedException e ) { }
            
            IpAddress addr = services.getIpAddress(testAddress);
            
            assertNotNull("Address did not exist: " + testAddress, addr);
            assertEquals("Server assignment does not match", vmToKill, addr.getServerId());
        }
        else {
            try {
                services.assign(testAddress == null ? "99" : testAddress, vmToKill == null ? "99" : vmToKill);
                fail("Assigning private IPs is supposedly not supported, but the operation returned successfully with: " + addressToRelease);
            }
            catch( OperationNotSupportedException success ) {
                // success
            }            
        }
    }
    
    @Test
    public void testAssignPublicIp() throws CloudException, InternalException {
        AddressServices services = provider.getAddressServices();
        
        if( services.isAssigned(AddressType.PUBLIC) ) {
            services.assign(testAddress, vmToKill);
            try { Thread.sleep(5000L); }
            catch( InterruptedException e ) { }
            
            IpAddress addr = services.getIpAddress(testAddress);
            
            assertNotNull("Address did not exist: " + testAddress, addr);
            assertEquals("Server assignment does not match", vmToKill, addr.getServerId());
        }
        else {
            try {
                services.assign(testAddress == null ? "99" : testAddress, vmToKill == null ? "99" : vmToKill);
                fail("Assigning public IPs is supposedly not supported, but the operation returned successfully with: " + addressToRelease);
            }
            catch( OperationNotSupportedException success ) {
                // success
            }            
        }
    }
    
    @Test
    public void testForward() throws CloudException, InternalException {
        AddressServices services = provider.getAddressServices();
        
        if( services.isForwarding() ) {
            services.forward(testAddress, 9321, Protocol.TCP, 22, vmToKill);
            try { Thread.sleep(5000L); }
            catch( InterruptedException e ) { }

            boolean found = false;
            
            for( IpForwardingRule rule : services.listRules(testAddress) ) {
                if( rule.getPrivatePort() != 22 ) {
                    continue;
                }
                if( rule.getPublicPort() != 9321 ) {
                    continue;
                }
                if( !rule.getServerId().equals(vmToKill) ) {
                    continue;
                }
                if( !rule.getProtocol().equals(vmToKill) ) {
                    continue;
                }
                if( !rule.getAddressId().equals(testAddress) ) {
                    continue;
                }
                found = true;
                break;
            }
            assertTrue("Rule silently failed (no rule listed)", found);
        }
        else {
            try {
                services.forward(testAddress == null ? "99" : testAddress, 9321, Protocol.TCP, 22, vmToKill == null ? "99" : vmToKill);
                fail("Assigning public IPs is supposedly not supported, but the operation returned successfully with: " + addressToRelease);
            }
            catch( OperationNotSupportedException success ) {
                // success
            }            
        }
    }
    
    @Test
    public void testGetAddress() throws CloudException, InternalException {
        if( testAddress == null ) {
            return;
        }
        AddressServices services = provider.getAddressServices();

        IpAddress addr = services.getIpAddress(testAddress);
        
        assertNotNull("Address not found: " + testAddress, addr);
        assertEquals("Address does not match target", testAddress, addr.getProviderIpAddressId());
    }
    
    @Test
    public void testGetBogusAddress() throws CloudException, InternalException {
        AddressServices services = provider.getAddressServices();

        IpAddress addr = services.getIpAddress(UUID.randomUUID().toString());
        
        assertNull("Found address at: " + testAddress, addr);
    }
    
    @Test
    public void testListAllPublicAddresses() throws CloudException, InternalException {
        assertNotNull("No address list returned", provider.getAddressServices().listPublicIpPool(false));
    }
    
    @Test 
    public void testListAllPrivateAddresses() throws CloudException, InternalException {
        assertNotNull("No address list returned", provider.getAddressServices().listPrivateIpPool(false));
    }
    
    @Test
    public void testListAvailablePublicAddresses() throws CloudException, InternalException {
        Iterable<IpAddress> list = provider.getAddressServices().listPublicIpPool(true);
        
        assertNotNull("No address list returned", list);
        for( IpAddress address : list ) {
            assertNull("No load balancer value should exist for an unassigned IP (" + address.getProviderLoadBalancerId() + ")", address.getProviderLoadBalancerId());
            assertNull("No server value should exist for an unassigned IP (" + address.getServerId() + ")", address.getServerId());
        }
    }
    
    @Test 
    public void testListAvailablePrivateAddresses() throws CloudException, InternalException {
        Iterable<IpAddress> list = provider.getAddressServices().listPrivateIpPool(true);

        assertNotNull("No address list returned", list);
        for( IpAddress address : list ) {
            assertNull("No load balancer value should exist for an unassigned IP (" + address.getProviderLoadBalancerId() + ")", address.getProviderLoadBalancerId());
            assertNull("No server value should exist for an unassigned IP (" + address.getServerId() + ")", address.getServerId());
        }
    }
    
    @Test
    public void testListForwardRules() throws CloudException, InternalException {
        Iterable<IpForwardingRule> rules = provider.getAddressServices().listRules(testAddress);
        
        assertNotNull("No rule list returned", rules);
        for( IpForwardingRule rule : rules ) {
            assertNotNull("Rule does not have an ID", rule.getProviderRuleId());
            assertNotNull("Rule does not specify a protocol", rule.getProtocol());
            assertEquals("Address does not match rule", testAddress, rule.getAddressId());
        }
    }
    
    @Test
    public void testMetaData() throws CloudException, InternalException {
        AddressServices services = provider.getAddressServices();

        services.isAssigned(AddressType.PUBLIC);
        services.isAssigned(AddressType.PRIVATE);
        services.isForwarding();
        services.isRequestable(AddressType.PUBLIC);
        services.isRequestable(AddressType.PRIVATE);
    }
    
    @Test
    public void testReleasePrivateIpFromPool() throws CloudException, InternalException {
        AddressServices services = provider.getAddressServices();
        
        if( services.isRequestable(AddressType.PRIVATE) ) {
            services.releaseFromPool(addressToRelease);
            try { Thread.sleep(5000L); }
            catch( InterruptedException e ) { }
            IpAddress tmp = services.getIpAddress(addressToRelease);
            
            assertNull("Address still exists after release: " + addressToRelease, tmp);
            addressToRelease = null;
        }
        else {
            try {
                services.releaseFromPool(addressToRelease == null ? "99" : addressToRelease);
                fail("Requesting private IPs is supposedly not supported, but the operation returned successfully with: " + addressToRelease);
            }
            catch( OperationNotSupportedException success ) {
                // success
            }
        }
    }
    
    @Test
    public void testReleasePrivateIpFromServer() throws CloudException, InternalException {
        AddressServices services = provider.getAddressServices();
        
        if( services.isAssigned(AddressType.PRIVATE) ) {
            try {
                services.releaseFromServer(testAddress);
                try { Thread.sleep(5000L); }
                catch( InterruptedException e ) { }
                IpAddress tmp = services.getIpAddress(testAddress);
                
                assertTrue("Address is still attached: " + testAddress, tmp.getServerId() == null || !tmp.getServerId().equals(vmToKill));
            }
            catch( OperationNotSupportedException e ) {
                // some clouds may not support releasing a private IP
            }
        }
        else {
            try {
                services.releaseFromServer(testAddress == null ? "99" : testAddress);
                fail("Assigning private IPs is supposedly not supported, but the operation returned successfully with: " + addressToRelease);
            }
            catch( OperationNotSupportedException success ) {
                // success
            }
        }
    }
    
    @Test
    public void testReleasePublicIpFromPool() throws CloudException, InternalException {
        AddressServices services = provider.getAddressServices();
        
        if( services.isRequestable(AddressType.PUBLIC) ) {
            services.releaseFromPool(addressToRelease);
            try { Thread.sleep(5000L); }
            catch( InterruptedException e ) { }
            IpAddress tmp = services.getIpAddress(addressToRelease);
            
            assertNull("Address still exists after release: " + addressToRelease, tmp);
            addressToRelease = null;
        }
        else {
            try {
                services.releaseFromPool(addressToRelease == null ? "99" : addressToRelease);
                fail("Requesting private IPs is supposedly not supported, but the operation returned successfully with: " + addressToRelease);
            }
            catch( OperationNotSupportedException success ) {
                // success
            }
        }
    }
    
    @Test
    public void testReleasePublicIpFromServer() throws CloudException, InternalException {
        AddressServices services = provider.getAddressServices();
        
        if( services.isAssigned(AddressType.PUBLIC) ) {
            services.releaseFromServer(testAddress);
            try { Thread.sleep(5000L); }
            catch( InterruptedException e ) { }
            IpAddress tmp = services.getIpAddress(testAddress);
            
            assertTrue("Address is still attached: " + testAddress, tmp.getServerId() == null || !tmp.getServerId().equals(vmToKill));
        }
        else {
            try {
                services.releaseFromServer(testAddress == null ? "99" : testAddress);
                fail("Assigning private IPs is supposedly not supported, but the operation returned successfully with: " + addressToRelease);
            }
            catch( OperationNotSupportedException success ) {
                // success
            }
        }
    }
    
    @Test
    public void testRequestPrivateIp() throws CloudException, InternalException {
        AddressServices services = provider.getAddressServices();
        
        if( services.isRequestable(AddressType.PRIVATE) ) {
            addressToRelease = services.request(AddressType.PRIVATE);
            assertNotNull("No address resulted from your allocation request", addressToRelease);
        }
        else {
            try {
                addressToRelease = services.request(AddressType.PRIVATE);
                fail("Requesting private IPs is supposedly not supported, but the operation returned successfully with: " + addressToRelease);
            }
            catch( OperationNotSupportedException success ) {
                // success
            }
        }
    }
    
    @Test
    public void testRequestPublicIp() throws CloudException, InternalException {
        AddressServices services = provider.getAddressServices();
        String addressId;
        
        if( services.isRequestable(AddressType.PUBLIC) ) {
            addressId = services.request(AddressType.PUBLIC);
            assertNotNull("No address resulted from your allocation request", addressId);
        }
        else {
            try {
                addressId = services.request(AddressType.PUBLIC);
                fail("Requesting public IPs is supposedly not supported, but the operation returned successfully with: " + addressId);
            }
            catch( OperationNotSupportedException success ) {
                // success
            }
        }
    }
    
    @Test
    public void testStopBogusForward() throws CloudException, InternalException {
        if( provider.getAddressServices().isForwarding() ) {
            boolean success = false;
            
            try {
                provider.getAddressServices().stopForward("-2");
                success = true;
            }
            catch( Throwable t ) {
                // desired outcome
            }
            assertTrue("Attempt to stop forwarding for bogus rule incorrectly succeeded", !success);
        }
        else {
            try {
                provider.getAddressServices().stopForward("-2");
                fail("Attempts to stop forwarding should throw OperationNotSupportedException when forwarding is not supported");
            }
            catch( OperationNotSupportedException e ) {
                // success
            }
        }
    }
    
    @Test
    public void testStopForward() throws CloudException, InternalException {
        if( provider.getAddressServices().isForwarding() ) {
            provider.getAddressServices().stopForward(testRuleId);
            boolean found = false;
            
            for( IpForwardingRule rule : provider.getAddressServices().listRules(testAddress) ) {
                if( !rule.getAddressId().equals(testAddress) ) {
                    continue;
                }
                if( rule.getPrivatePort() != 22 ) {
                    continue;
                }
                if( rule.getPublicPort() != 9321 ) {
                    continue;
                }
                if( !rule.getProtocol().equals(Protocol.TCP) ) {
                    continue;
                }
                if( rule.getServerId() == null || !rule.getServerId().equals(vmToKill) ) {
                    continue;
                }
                found = true;
                break;
            }
            assertTrue("The test rule is still in place", !found);
            testRuleId = null;
        }
        else {
            try {
                provider.getAddressServices().stopForward("-2");
                fail("Attempts to stop forwarding should throw OperationNotSupportedException when forwarding is not supported");
            }
            catch( OperationNotSupportedException e ) {
                // success
            }
        }
    }
    
    @Test
    public void testSubscribed() throws CloudException, InternalException {
        AddressServices services = provider.getAddressServices();

        assertTrue("No subscription exists for this service", services.isSubscribed());
    }
}
